
.md-header__button.md-logo :where(img,svg) {
    width: 100%;
    height: 30px;
}

.md-header, .md-header__inner {
    background-color: #fff;
    color: #2A4CDF;
}

.md-header[data-md-state=shadow] {
    box-shadow: 0px 2px 4px -1px rgb(0 0 0 / 20%), 0px 4px 5px 0px rgb(0 0 0 / 14%), 0px 1px 10px 0px rgb(0 0 0 / 12%);
    transition: box-shadow 200ms linear;
}

.md-header__inner {
    height: 80px;
}

.md-header__topic {
    font-size: 24px;
}

.md-footer {
    background-color: #2A4CDF;
}

.md-search__form:hover {
    background-color: rgba(0,0,0,.13);
}

.md-typeset h1 {
    color: black;
    font-weight: 700;
}

.youtube {
    position: relative;
    width: 100%;
    height: 0;
    padding-bottom: 56.25%;
}

.youtube iframe {
    position: absolute;
    top: 0;
    left: 0;
    width: 100%;
    height: 100%;
}

.home-grid {
    display: grid;
    grid-template-columns: 1fr 1fr;
    grid-column-gap: 24px;
    row-gap: 24px;
}

.home-grid div {
    border-radius: 4px;
    padding: 24px;
    background-color: #f3e9d9;
}

.home-grid h3 {
    margin-top: 0;
    font-weight: 700;
}